from .dataset import (
    AbstractOccInstDatasetMeta,
    NiftiOccInstDatasetMeta,
    DescriptiveMetaData,
    PlantNiftiOccInstDatasetMeta,
)
from .data_splitting import AbstractDataSplitter, LocalDataSplitter, MinioDataSplitter
from .data_module import (
    AbstractDataModule3dImgSeg,
    LocalDataModule3DImgSeg,
    MinioDataModule3DImgSeg,
)
from .json_helpers import load_from_json_file, save_to_json_file

__all__ = [
    "AbstractOccInstDatasetMeta",
    "NiftiOccInstDatasetMeta",
    "DescriptiveMetaData",
    "PlantNiftiOccInstDatasetMeta",
    "AbstractDataSplitter",
    "LocalDataSplitter",
    "MinioDataSplitter",
    "AbstractDataModule3dImgSeg",
    "LocalDataModule3DImgSeg",
    "MinioDataModule3DImgSeg",
    "load_from_json_file",
    "save_to_json_file",
]



from typing import List, Dict
import copy
import numpy as np
from nibabel import Nifti1Image

from magdi_data.dataset.abstract_occ_inst_dataset_meta import AbstractOccInstDatasetMeta
import nibabel as nib


class NiftiOccInstDatasetMeta(AbstractOccInstDatasetMeta):

    def update_and_validate_based_on_nifti_headers(self, dataset_path):
        instance_list: List[Dict[str, str]] = (
            self.get_data_split_paths_and_nifti_headers(dataset_path, "training")
        )
        instance_list_test: List[Dict[str, str]] = (
            self.get_data_split_paths_and_nifti_headers(dataset_path, "test")
        )
        instance_list_validation: List[Dict[str, str]] = (
            self.get_data_split_paths_and_nifti_headers(dataset_path, "validation")
        )
        instance_list += instance_list_test
        instance_list += instance_list_validation

        if len(instance_list) == 0:
            raise ValueError(
                "Tried to update and validate fields based on Nifti "
                "headers but no instances have been found."
            )

        self._validate_image_dtype(instance_list)
        self._validate_annotation_dtype(instance_list)
        self._update_and_validate_dimensions(instance_list)

    def _validate_annotation_dtype(self, instance_list: List[Dict[str, str | dict]]):
        img_dtype_dict: dict = {}

        for instance_dict in instance_list:
            if "annotations_header" in instance_dict:
                header: dict = instance_dict["annotations_header"]  # type: ignore
                dtype: str = str(header["datatype"].dtype)

                img_dtype_dict[instance_dict["image"]] = dtype

        img_dtype_list = list(img_dtype_dict.values())
        if len(img_dtype_list) == 0:
            return
        dtype0 = img_dtype_list[0]
        for path, dtype in img_dtype_dict.items():
            if dtype0 != dtype:
                raise ValueError(
                    f"Image {path} has a different datatype ({dtype}) "
                    f"than the first instance ({dtype0})."
                )

        if dtype0 != self.image_data_type:
            raise ValueError(
                f"A different image_data_type ({self.image_data_type}) "
                f"is set than the first instance has ({dtype0})."
            )

    def _validate_image_dtype(self, instance_list: List[Dict[str, str | dict]]):
        img_dtype_dict: dict = {}

        for instance_dict in instance_list:

            header: dict = instance_dict["image_header"]  # type: ignore
            dtype: str = str(header["datatype"].dtype)

            img_dtype_dict[instance_dict["image"]] = dtype

        dtype0 = list(img_dtype_dict.values())[0]
        for path, dtype in img_dtype_dict.items():
            if dtype0 != dtype:
                raise ValueError(
                    f"Image {path} has a different datatype ({dtype}) "
                    f"than the first instance ({dtype0})."
                )

        if dtype0 != self.image_data_type:
            raise ValueError(
                f"A different image_data_type ({self.image_data_type}) "
                f"is set than the first instance has ({dtype0})."
            )

    def _update_and_validate_dimensions(
        self, instance_list: List[Dict[str, str | dict]]
    ):
        """
        Update dimension metadata fields based on the Nifti header of each data
        instance.
        Validate if image dimensions fit to annotation dimensions if annotation
        dimensions exist.
        """

        img_dimensions_list: List[List[int]] = []
        img_voxel_sizes_list: List[List[int]] = []

        range_initialized: bool = False
        range_min: int = 0
        range_max: int = 0

        range_avg_values: list[float] = []

        for instance_dict in instance_list:

            img_header: dict = instance_dict["image_header"]  # type: ignore

            img_3d_dimensions = img_header["dim"][1:4]
            img_3d_voxel_sizes = img_header["pixdim"][1:4]
            if not range_initialized:
                range_min = img_header["glmin"]
            else:
                range_min = min(range_min, img_header["glmin"])
            if not range_initialized:
                range_max = img_header["glmax"]
            else:
                range_max = min(range_max, img_header["glmax"])
            range_initialized = True
            img_range_avg = img_header["range_avg"]

            if "annotations_header" in instance_dict:
                ann_header: dict = instance_dict["annotations_header"]  # type: ignore

                ann_3d_dimensions = ann_header["dim"][1:4]
                ann_3d_voxel_sizes = img_header["pixdim"][1:4]

                if not np.array_equal(img_3d_dimensions, ann_3d_dimensions):
                    raise ValueError(
                        "Annotation dimensions do not fit with the image dimensions "
                        "(header['dim'][1:4])"
                        f"for {instance_dict['image']} and "
                        f"{instance_dict['annotations']}."
                    )
                if not np.array_equal(img_3d_voxel_sizes, ann_3d_voxel_sizes):
                    raise ValueError(
                        "Annotation voxel sizes do not fit with the image voxel sizes "
                        "(header['pixdim'][1:4])"
                        f"for {instance_dict['image']} and "
                        f"{instance_dict['annotations']}."
                    )

            img_dimensions_list.append(img_3d_dimensions)
            img_voxel_sizes_list.append(img_3d_voxel_sizes)
            range_avg_values.append(img_range_avg)

        if len(img_dimensions_list) == 0:
            return

        self.range_min = int(range_min)
        self.range_max = int(range_max)
        self.range_avg = float(np.mean(range_avg_values))

        self.dimensions_max, self.dimensions_min, self.dimensions_avg = (
            self._get_shape_statistics(img_dimensions_list)
        )

        (
            self.resolution_voxel_size_max,
            self.resolution_voxel_size_min,
            self.resolution_voxel_size_avg,
        ) = self._get_shape_statistics(img_voxel_sizes_list)

    def get_data_split_paths_and_nifti_headers(
        self, dataset_path: str, split: str = "training"
    ) -> List[Dict[str, str | dict]]:
        """
        Extends method get_data_split_paths().
        Get nifti headers directly from the nifti files.
        Make a deep copy of the path dictionaries of the data instances.
        Add the nifti headers to the dictionaries of each instance.

        It is an expensive operation, because each nifti file is loaded to extract the
        header. This way, a lot of metadata is read for further use.
        """

        file_path_dict_list: List[Dict[str, str | dict]] = copy.deepcopy(
            self.get_data_split_paths(dataset_path, split)  # type: ignore
        )

        for file_path_dict in file_path_dict_list:
            if "image" in file_path_dict:
                nii: Nifti1Image = nib.load(file_path_dict["image"])  # type: ignore
                nifti_header: dict = dict(nii.header)  # type: ignore
                image_data = nii.get_fdata()
                global_min, global_max = np.min(image_data), np.max(image_data)

                # https://www.nitrc.org/docman/view.php/26/64/nifti1.h
                # glmax and glmin are specified but unused fields inside the NIFTI
                # header.
                # They are for compatibility with analyze
                # We use both fields here to temporary store the max and min values.
                nifti_header["glmax"] = global_max
                nifti_header["glmin"] = global_min

                # additional temporary field
                nifti_header["range_avg"] = np.mean(image_data)

            else:
                raise FileNotFoundError(
                    "Image dimensions could not be read, "
                    "because no file path for key 'image' is set."
                )

            file_path_dict["image_header"] = nifti_header

            if "annotations" in file_path_dict:
                ann_header = dict(
                    nib.load(file_path_dict["annotations"]).header  # type: ignore
                )
                file_path_dict["annotations_header"] = ann_header

        return file_path_dict_list



"""
Contains a subclass of AbstractDataModule3dImgSeg to load data from local.
"""

from magdi_data.data_module.abstract_data_module_3d_img import (
    AbstractDataModule3dImgSeg,
)
import json
from pydantic_core import from_json
from monai.data import CacheDataset
from magdi_data.dataset.nifti_occ_inst_dataset_meta import NiftiOccInstDatasetMeta
from magdi_data.json_helpers import load_from_json_file


class LocalDataModule3DImgSeg(AbstractDataModule3dImgSeg):

    def __init__(
        self,
        dataset_path: str,
        roi_size,
        batch_size: int,
        num_workers: int,
        persistent_workers: bool,
        pin_memory: bool,
        custom_dataset_meta_path: str | None = None,
        **kwargs,
    ):

        super().__init__(
            roi_size=roi_size,
            batch_size=batch_size,
            num_workers=num_workers,
            persistent_workers=persistent_workers,
            pin_memory=pin_memory,
            **kwargs,
        )

        self.dataset_path = dataset_path
        self.custom_dataset_meta_path = custom_dataset_meta_path

    def prepare_data(self) -> None:
        print("Preparing data")
        dataset_meta_path = self.get_dataset_meta_path(
            self.dataset_path, self.custom_dataset_meta_path
        )
        self.dataset_meta = NiftiOccInstDatasetMeta.model_validate(
            from_json(json.dumps(load_from_json_file(dataset_meta_path)))
        )
        print(self.dataset_meta)

    def setup(self, stage: str):
        super().setup(stage=stage)  # important
        if stage == "fit":
            # CacheDataset caches non-random transforms
            training_dataset_paths = self.dataset_meta.get_data_split_paths(
                self.dataset_path, split="training"
            )
            self.train_dataset = CacheDataset(
                data=training_dataset_paths,
                transform=self.get_train_transforms(),
                num_workers=self.num_workers,
                cache_rate=self.cache_rate,
                cache_num=self.cache_num,
            )

            val_dataset_paths = []
            if self.dataset_meta.validation is not None:
                val_dataset_paths = self.dataset_meta.get_data_split_paths(
                    self.dataset_path, split="validation"
                )
            self.val_dataset = CacheDataset(
                data=val_dataset_paths,
                transform=self.get_val_transforms(),
                num_workers=self.num_workers,
                cache_rate=self.cache_rate,
                cache_num=self.cache_num,
            )

        elif stage == "test":
            test_dataset_paths = self.dataset_meta.get_data_split_paths(
                self.dataset_path, split="test"
            )
            self.test_dataset = CacheDataset(
                data=test_dataset_paths,
                transform=self.get_test_transforms(),
                num_workers=self.num_workers,
                cache_rate=self.cache_rate,
                cache_num=self.cache_num,
            )



from .plant_nifti_occ_inst_dataset_meta import (
    DescriptiveMetaData,
    PlantNiftiOccInstDatasetMeta,
)
from .nifti_occ_inst_dataset_meta import NiftiOccInstDatasetMeta
from .abstract_occ_inst_dataset_meta import AbstractOccInstDatasetMeta

__all__ = [
    "AbstractOccInstDatasetMeta",
    "NiftiOccInstDatasetMeta",
    "DescriptiveMetaData",
    "PlantNiftiOccInstDatasetMeta",
]



from .abstract_data_splitter import AbstractDataSplitter
from .local_data_splitter import LocalDataSplitter
from .minio_data_splitter import MinioDataSplitter

__all__ = ["AbstractDataSplitter", "LocalDataSplitter", "MinioDataSplitter"]



"""
Contains a subclass of AbstractDataSplitter to split a dataset whose metadata file is
located locally.
"""

from typing_extensions import override
from pydantic_core import from_json
from magdi_data.data_splitting.abstract_data_splitter import AbstractDataSplitter
from magdi_data.dataset.abstract_occ_inst_dataset_meta import AbstractOccInstDatasetMeta


class LocalDataSplitter(AbstractDataSplitter):
    def __init__(
        self,
        local_dataset_path: str,
        custom_dataset_meta_path: str | None = None,
    ):
        self.dataset_path = local_dataset_path
        self.dataset_meta_path = self.validate_dataset_meta_path(
            self.dataset_path, custom_dataset_meta_path
        )

    @override
    def load_dataset_meta_file(self):
        """
        Read the addressed dataset meta JSON file on disk to create an instance of
        AbstractOccInstDatasetMeta.

        Returns:
            Instance of AbstractOccInstDatasetMeta
        """

        with open(self.dataset_meta_path, mode="r", encoding="utf-8") as file:
            json_str_input = file.read()

        return AbstractOccInstDatasetMeta.model_validate(from_json(json_str_input))



"""
Contains helper functions.
"""

import json
import os


def load_from_json_file(path):
    with open(path, mode="r", encoding="utf-8") as file:
        json_str_input = file.read()
    return json.loads(json_str_input)


def save_to_json_file(data, filepath: str, sort_keys=False):
    json_str = json.dumps(data, indent=4, sort_keys=sort_keys)
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    with open(filepath, mode="w", encoding="utf-8") as text_file:
        text_file.write(json_str)



from pydantic import Field, BaseModel

from magdi_data.dataset.nifti_occ_inst_dataset_meta import NiftiOccInstDatasetMeta


class DescriptiveMetaData(BaseModel):
    latin_name: str = Field(default="")
    line_name: str = Field(default="")
    structure: str = Field(default="")
    dap: str = Field(default="")
    device: str = Field(default="")
    coil: str = Field(default="")
    measurement_channel: str = Field(default="")


class PlantNiftiOccInstDatasetMeta(NiftiOccInstDatasetMeta):

    descriptive_metadata: DescriptiveMetaData = Field(
        default_factory=DescriptiveMetaData
    )



import os
import re
from datetime import date

import numpy as np
from pydantic import BaseModel, Field, field_validator, model_validator
from typing import List, Dict, Optional, Any, Self
import copy
import uuid


class AbstractOccInstDatasetMeta(BaseModel):
    """
    Abstract model for the metadata file of a dataset that has the Occurrence Instance
    Dataset Structure.
    """

    uuid: str = Field(default=str(uuid.uuid4()))  # is set automatically if not set
    name: str = Field(default="")
    short_name: str = Field(default="")  # equal to name if not set manually
    release: str = Field(default=date.today().strftime("%Y.%m.%d"))
    id: str = Field(default="")  # is set automatically
    descriptive_metadata: Any = Field(default={})
    entity: str = Field(default="")
    description: str = Field(default="")
    reference: str = Field(default="")
    license: str = Field(default="")
    tags: List[str] = Field(default=[])
    task_categories: List[str] = Field(default=[])
    labels: Optional[Dict[str, str]] = Field(default=None)
    image_format: str = Field(default="")
    annotation_format: Optional[str] = Field(default=None)
    num_instances_total: int = Field(default=0)
    num_instances_training: int = Field(default=0)
    num_instances_test: Optional[int] = Field(default=None)
    num_instances_validation: Optional[int] = Field(default=None)
    num_occurrences_total: int = Field(default=0)
    num_occurrences_training: int = Field(default=0)
    num_occurrences_test: Optional[int] = Field(default=None)
    num_occurrences_validation: Optional[int] = Field(default=None)
    image_data_type: str = Field(default="int16")
    annotation_data_type: Optional[str] = Field(default=None)
    value_channels: int = Field(default=1)
    range: List[int] = Field(default=[0, 0])
    range_max: int = Field(default=0)
    range_min: int = Field(default=0)
    range_avg: float = Field(default=0)
    dimensions_max: Optional[List[int]] = Field(default=None)
    dimensions_min: Optional[List[int]] = Field(default=None)
    dimensions_avg: Optional[List[float]] = Field(default=None)
    resolution_unit: str = Field(default="mm")
    resolution_voxel_size: List[float] = Field(default=[1, 1, 1])
    resolution_voxel_size_max: Optional[List[float]] = Field(default=None)
    resolution_voxel_size_min: Optional[List[float]] = Field(default=None)
    resolution_voxel_size_avg: Optional[List[float]] = Field(default=None)
    training: List[Dict[str, str]] = Field(default=[])
    test: Optional[List[Dict[str, str]]] = Field(default=None)
    validation: Optional[List[Dict[str, str]]] = Field(default=None)
    data_references: Optional[Dict[str, str]] = Field(default={})

    @field_validator("uuid", mode="after")
    def set_uuid(cls, v) -> str:
        # Check if the value is not a valid UUID, then overwrite
        # Raise value error if v is not valid
        return str(uuid.UUID(v))

    @staticmethod
    def _transform_id_string(name: str, release: str, uuid: str) -> str:
        lower_str = (name + "_" + release.replace(".", "-")).lower()
        transformed_string = re.sub(r"[^a-z0-9-]", "_", lower_str) + "_" + uuid
        return transformed_string

    @property
    def get_id(self) -> str:
        return self._transform_id_string(self.name, self.release, str(self.uuid))

    @field_validator("short_name", mode="after")
    def validate_short_name(cls, v, values) -> str:
        if not v:
            return values.data["name"]
        return v

    @field_validator("id", mode="after")
    def validate_id(cls, v, values) -> str:
        return cls._transform_id_string(
            values.data["name"], values.data["release"], str(values.data["uuid"])
        )

    @model_validator(mode="after")
    def validate_id_after(self) -> Self:
        self.id = self._transform_id_string(self.name, self.release, str(self.uuid))
        return self

    @field_validator("image_data_type", "annotation_data_type")
    def validate_np_dtype(cls, v):
        assert isinstance(np.dtype(v), np.dtype)
        return v

    @model_validator(mode="after")
    def validate_fields_based_on_data_split(self):
        """
        Update metadata fields based on the training, validation, and test split
        """
        self.update_counters()
        return self

    @staticmethod
    def count_occurrences_and_instances(data, key="image"):
        occurrence_dict: dict = {}

        for item in data:
            image_path = item[key]
            parts = image_path.split("/")
            occurrence = parts[0]  # 'occurrence-xxxx'
            instance = parts[1]  # 'instance-xxxx'

            if occurrence not in occurrence_dict:
                occurrence_dict[occurrence] = set()
            occurrence_dict[occurrence].add(instance)

        total_occurrences = len(occurrence_dict)
        total_instances = sum(len(instances) for instances in occurrence_dict.values())

        return total_occurrences, total_instances

    def update_counters(self):
        self.num_occurrences_training, self.num_instances_training = (
            self.count_occurrences_and_instances(self.training)
        )
        self.num_occurrences_total = self.num_occurrences_training
        self.num_instances_total = self.num_instances_training
        if self.test is not None:
            self.num_occurrences_test, self.num_instances_test = (
                self.count_occurrences_and_instances(self.test)
            )
            self.num_occurrences_total += self.num_occurrences_test
            self.num_instances_total += self.num_instances_test
        else:
            self.num_occurrences_test = None
            self.num_instances_test = None

        if self.validation is not None:
            self.num_occurrences_validation, self.num_instances_validation = (
                self.count_occurrences_and_instances(self.validation)
            )
            self.num_occurrences_total += self.num_occurrences_validation
            self.num_instances_total += self.num_instances_validation
        else:
            self.num_occurrences_validation = None
            self.num_instances_validation = None

    def update_dimensions(self, dataset_path: str):
        shapes = self.get_dimensions_in_split(dataset_path, "training")

        if self.test is not None:
            shapes += self.get_dimensions_in_split(dataset_path, "test")
        if self.validation is not None:
            shapes += self.get_dimensions_in_split(dataset_path, "validation")

        if len(shapes) == 0:
            return

        self.dimensions_max, self.dimensions_min, self.dimensions_avg = (
            self._get_shape_statistics(shapes)
        )

    @staticmethod
    def _get_shape_statistics(shapes):
        np_shapes = np.array(shapes)
        min_shape = list(np_shapes.min(axis=0))
        max_shape = list(np_shapes.max(axis=0))
        mean_shape = list(np_shapes.mean(axis=0))
        return max_shape, min_shape, mean_shape

    def get_dimensions_in_split(
        self, dataset_path: str, split: str = "training"
    ) -> List[tuple[int]]:
        return []

    def get_annotations_dtype(self) -> np.dtype:
        return np.dtype(self.annotation_data_type)

    def get_image_dtype(self) -> np.dtype:
        return np.dtype(self.image_data_type)

    def get_data_split_paths(
        self, dataset_path: str, split: str = "training"
    ) -> List[Dict[str, str]]:
        return self.concat_paths(dataset_path, split)

    def concat_paths(
        self, dataset_path: str, split: str = "training"
    ) -> List[Dict[str, str]]:
        instance_list = copy.deepcopy(getattr(self, split))
        if instance_list is None:
            return []

        for file_path_dict in instance_list:
            for key in file_path_dict:
                file_path_dict[key] = os.path.join(dataset_path, file_path_dict[key])
        return instance_list



from .abstract_data_module_3d_img import AbstractDataModule3dImgSeg
from .local_data_module_3d_img import LocalDataModule3DImgSeg
from .minio_data_module_3d_img import MinioDataModule3DImgSeg

__all__ = [
    "AbstractDataModule3dImgSeg",
    "LocalDataModule3DImgSeg",
    "MinioDataModule3DImgSeg",
]



"""
Contains the implementation of a Lighting Data module to provide data for 3D
image segmentation.
"""

import os
from typing import Callable, Optional, Dict
import os.path
import lightning as L
from monai.data import DataLoader, list_data_collate
from monai.transforms import (
    Compose,
    RandSpatialCropSamplesd,
    Transform,
)
from torch.utils.data import Dataset
from typing_extensions import override
from magdi_data.dataset.nifti_occ_inst_dataset_meta import NiftiOccInstDatasetMeta
from transformers import ImageProcessingMixin


def default_augmentations() -> Transform | None:
    return None


class AbstractDataModule3dImgSeg(L.LightningDataModule):

    def __init__(
        self,
        roi_size,
        batch_size: int,
        num_workers: int,
        persistent_workers: bool,
        pin_memory: bool,
        **kwargs,
    ):

        super().__init__()
        self.batch_size = batch_size
        self.roi_size = roi_size

        self.num_workers = num_workers
        self.persistent_workers = persistent_workers
        self.pin_memory = pin_memory

        self.cache_rate: float = 0.0  # 1.0
        self.cache_num: int = 0  # sys.maxsize

        self.train_dataset: Dataset | None = None
        self.val_dataset: Dataset | None = None
        self.test_dataset: Dataset | None = None

        self.augmentation_hook: Callable[[], Transform | None] = default_augmentations

        self.dataset_meta: NiftiOccInstDatasetMeta = None  # type: ignore

        # Specific processor implementation for a Hugging Face model
        self.hf_processor: ImageProcessingMixin = None  # type: ignore

    def get_dataset_meta_path(self, dataset_path, custom_dataset_meta_path) -> str:
        if custom_dataset_meta_path is None:
            return os.path.join(dataset_path, "dataset_meta.json")
        else:
            return custom_dataset_meta_path

    def get_labels(self) -> Optional[Dict[str, str]]:
        if self.dataset_meta is None:
            raise ValueError("self.dataset_meta is None")
        return self.dataset_meta.labels

    def get_test_transforms(self):
        return self.hf_processor

    def get_val_transforms(self):
        return self.hf_processor

    def get_train_transforms(self):

        train_transforms = [
            self.hf_processor,
            RandSpatialCropSamplesd(
                keys=["image", "annotations"],
                roi_size=self.roi_size,
                num_samples=1,
            ),
        ]

        augmentation_transforms = self.augmentation_hook()
        if augmentation_transforms is not None:
            train_transforms.append(augmentation_transforms)  # type: ignore

        return Compose(train_transforms)

    @override
    def setup(self, stage: str):
        if self.dataset_meta is None:
            raise ValueError(
                "self.dataset_meta is not allowed to be None."
                "Set it before calling setup()"
            )
        if self.hf_processor is None:
            raise ValueError(
                "self.hf_processor is not allowed to be None."
                "Set a Hugging Face processor before calling setup()."
            )

    @override
    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self.train_dataset,  # type: ignore
            shuffle=True,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=self.persistent_workers,
            pin_memory=self.pin_memory,
            collate_fn=list_data_collate,
        )

    @override
    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            self.val_dataset,  # type: ignore
            shuffle=False,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=self.persistent_workers,
            pin_memory=self.pin_memory,
            collate_fn=list_data_collate,
        )

    @override
    def test_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset,  # type: ignore
            shuffle=False,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=self.persistent_workers,
            pin_memory=self.pin_memory,
            collate_fn=list_data_collate,
        )
